#
# Copyright 2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""API provider for Multipass.

This implementation interfaces with multipass using the `multipass` command-line
utility.
"""

import io
import json
import locale
import logging
import pathlib
import shlex
import subprocess
import time
from typing import Any, Callable, Dict, List, Optional, Tuple

import packaging.version

from craft_providers import errors
from craft_providers.const import RETRY_WAIT

from .errors import MultipassError

logger = logging.getLogger(__name__)


class Multipass:
    """Wrapper for multipass command.

    :param multipass_path: Path to multipass command to use.
    :cvar minimum_required_version: Minimum required version for compatibility.
    """

    minimum_required_version = "1.7"

    def __init__(
        self, *, multipass_path: pathlib.Path = pathlib.Path("multipass")
    ) -> None:
        self.multipass_path = multipass_path

    def _run(self, command: List[str], **kwargs) -> subprocess.CompletedProcess:
        """Execute a multipass command.

        It always checks the result (as no errors should pass silently) and captures the
        output (so `multipass` does not pollute the terminal).
        """
        command = [str(self.multipass_path), *command]

        logger.debug("Executing on host: %s", shlex.join(command))
        return subprocess.run(command, check=True, capture_output=True, **kwargs)

    def delete(self, *, instance_name: str, purge=True) -> None:
        """Passthrough for running multipass delete.

        :param instance_name: The name of the instance_name to delete.
        :param purge: Flag to purge the instance_name's image after deleting.

        :raises MultipassError: on error.

        """
        command = ["delete", instance_name]
        if purge:
            command.append("--purge")

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to delete VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def exec(
        self,
        *,
        command: List[str],
        instance_name: str,
        runner: Callable = subprocess.run,
        timeout: Optional[float] = None,
        check: bool = False,
        **kwargs,
    ):
        """Execute command in instance_name with specified runner.

        The working directory the command is executed from inside the instance depends
        on the host's cwd. From the Multipass documentation:
        "In case we are executing the alias on the host from a directory which is
        mounted on the instance, the command will be executed on the instance from
        there. If the working directory is not mounted on the instance, the command will
        be executed on the default directory on the instance."

        :param command: Command to execute in the instance.
        :param instance_name: Name of instance to execute in.
        :param runner: Execution function to invoke, e.g. subprocess.run or
            Popen.  First argument is finalized command with the attached
            kwargs.
        :param timeout: Timeout (in seconds) for the command.
        :param check: Raise an exception if the command fails.
        :param kwargs: Additional kwargs for runner.

        :returns: Runner's instance.
        """
        final_cmd = [str(self.multipass_path), "exec", instance_name, "--", *command]

        quoted_final_cmd = shlex.join(final_cmd)
        logger.debug("Executing on host: %s", quoted_final_cmd)

        # Only subprocess.run supports timeout
        if runner is subprocess.run:
            return runner(final_cmd, timeout=timeout, check=check, **kwargs)

        return runner(final_cmd, **kwargs)

    def info(self, *, instance_name: str) -> Dict[str, Any]:
        """Get information/state for instance.

        :returns: Parsed json data from info command.

        :raises MultipassError: On error.
        """
        command = ["info", instance_name, "--format", "json"]

        try:
            proc = self._run(command, text=True)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to query info for VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

        return json.loads(proc.stdout)

    def is_supported_version(self) -> bool:
        """Check if Multipass version is supported.

        A helper to check if Multipass meets minimum supported version for
        craft-providers.

        :returns: True if installed version is supported.
        """
        minimum_version = packaging.version.parse(self.minimum_required_version)
        version, _ = self.version()

        return packaging.version.parse(version) >= minimum_version

    def launch(
        self,
        *,
        instance_name: str,
        image: str,
        cpus: Optional[str] = None,
        mem: Optional[str] = None,
        disk: Optional[str] = None,
    ) -> None:
        """Launch multipass VM.

        :param instance_name: The name the launched instance will have.
        :param image: Name of image to create the instance with.
        :param cpus: Amount of virtual CPUs to assign to the launched instance.
        :param mem: Amount of RAM to assign to the launched instance.
        :param disk: Amount of disk space the launched instance will have.

        :raises MultipassError: on error.
        """
        command = ["launch", image, "--name", instance_name]
        if cpus is not None:
            command.extend(["--cpus", cpus])
        if mem is not None:
            command.extend(["--memory", mem])
        if disk is not None:
            command.extend(["--disk", disk])

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to launch VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def list(self) -> List[str]:
        """List names of VMs.

        :returns: Data from stdout if instance exists, else None.

        :raises MultipassError: On error.
        """
        command = ["list", "--format", "json"]

        try:
            proc = self._run(command, text=True)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief="Failed to query list of VMs.",
                details=errors.details_from_called_process_error(error),
            ) from error

        data_list = json.loads(proc.stdout).get("list", [])
        return [instance["name"] for instance in data_list]

    def mount(
        self,
        *,
        source: pathlib.Path,
        target: str,
        uid_map: Optional[Dict[str, str]] = None,
        gid_map: Optional[Dict[str, str]] = None,
    ) -> None:
        """Mount host source path to target.

        :param source: Path of local directory to mount.
        :param target: Target mount points, in <name>[:<path>] format, where
            <name> is an instance name, and optional <path> is the mount point.
            If omitted, the mount point will be the same as the source's
            absolute path.
        :param uid_map: A mapping of user IDs for use in the mount of the form
            <host-id> -> <instance-id>.  File and folder ownership will be
            mapped from <host-id> to <instance-id> inside the instance.
        :param gid_map: A mapping of group IDs for use in the mount of the form
            <host-id> -> <instance-id>.  File and folder ownership will be
            mapped from <host-id> to <instance-id> inside the instance.
        """
        command = ["mount", str(source), target]

        if uid_map is not None:
            for host_id, instance_id in uid_map.items():
                command.extend(["--uid-map", f"{host_id}:{instance_id}"])

        if gid_map is not None:
            for host_id, instance_id in gid_map.items():
                command.extend(["--gid-map", f"{host_id}:{instance_id}"])

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to mount {str(source)!r} to {target!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def start(self, *, instance_name: str) -> None:
        """Start VM instance.

        :param instance_name: the name of the instance to start.

        :raises MultipassError: on error.
        """
        command = ["start", instance_name]

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to start VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def stop(self, *, instance_name: str, delay_mins: int = 0) -> None:
        """Stop VM instance.

        :param instance_name: the name of the instance_name to stop.
        :param delay_mins: Delay shutdown for specified number of minutes.

        :raises MultipassError: on error.
        """
        command = ["stop", instance_name]

        if delay_mins != 0:
            command.extend(["--time", str(delay_mins)])

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to stop VM {instance_name!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def transfer(self, *, source: str, destination: str) -> None:
        """Transfer to destination path with source IO.

        Multipass transfer uses sftp. By default, only the user `ubuntu` can transfer
        files. Therefore, the path inside the instance should be accessible by the
        `ubuntu` user.

        By default, Multipass only has access to the host's home directory. The host's
        path should be inside the home directory.

        :param source: The source path, prefixed with <name:> for a path inside
            the instance.
        :param destination: The destination path, prefixed with <name:> for a
            path inside the instance.

        :raises MultipassError: On error.
        """
        command = ["transfer", source, destination]

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to transfer {source!r} to {destination!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def transfer_destination_io(
        self, *, source: str, destination: io.BufferedIOBase, chunk_size: int = 4096
    ) -> None:
        """Transfer from source file to destination IO.

        Note that this can't use std{in,out}=open(...) due to LP #1849753.

        :param source: The source path, prefixed with <name:> for a path inside
            the instance.
        :param destination: An IO stream to write to.
        :param chunk_size: Number of bytes to transfer at a time.  Defaults to
            4096.

        :raises MultipassError: On error.
        """
        command = [str(self.multipass_path), "transfer", source, "-"]
        with subprocess.Popen(
            command,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:
            # Should never happen, but mypy/pyright makes noise.
            assert proc.stdout is not None
            assert proc.stderr is not None

            while True:
                data = proc.stdout.read(chunk_size)
                if not data:
                    break

                destination.write(data)

            # Take one read of stderr in case there is anything useful
            # for debugging an error.
            stderr = proc.stderr.read()

        if proc.returncode != 0:
            raise MultipassError(
                brief=f"Failed to transfer file {source!r}.",
                details=errors.details_from_command_error(
                    cmd=command, stderr=stderr, returncode=proc.returncode
                ),
            )

    def transfer_source_io(
        self, *, source: io.BufferedIOBase, destination: str, chunk_size: int = 4096
    ) -> None:
        """Transfer to destination path with source IO.

        Note that this can't use std{in,out}=open(...) due to LP #1849753.

        :param source: An IO stream to read from.
        :param destination: The destination path, prefixed with <name:> for a
            path inside the instance.
        :param chunk_size: Number of bytes to transfer at a time.  Defaults to
            4096.

        :raises MultipassError: On error.
        """
        command = [str(self.multipass_path), "transfer", "-", destination]
        with subprocess.Popen(
            command, stdin=subprocess.PIPE, stderr=subprocess.PIPE
        ) as proc:
            # Should never happen, but mypy/pyright makes noise.
            assert proc.stdin is not None
            assert proc.stderr is not None

            while True:
                data = source.read(chunk_size)
                if not data:
                    break

                proc.stdin.write(data)

            # Close stdin before reading stderr, otherwise read() will hang
            # because process is waiting for more data.
            proc.stdin.close()

            # Take one read of stderr in case there is anything useful
            # for debugging an error.
            stderr = proc.stderr.read()

        if proc.returncode != 0:
            raise MultipassError(
                brief=f"Failed to transfer file to destination {destination!r}.",
                details=errors.details_from_command_error(
                    cmd=command, stderr=stderr, returncode=proc.returncode
                ),
            )

    def umount(self, *, mount: str) -> None:
        """Unmount target in VM.

        :param mount: Mount point in <name>[:<path>] format, where <name> are
            instance names, and optional <path> are mount points.  If omitted,
            all mounts will be removed from the named instance.

        :raises MultipassError: On error.
        """
        command = ["umount", mount]

        try:
            self._run(command)
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief=f"Failed to unmount {mount!r}.",
                details=errors.details_from_called_process_error(error),
            ) from error

    def wait_until_ready(
        self, *, timeout: Optional[float] = None
    ) -> Tuple[str, Optional[str]]:
        """Wait until Multipass is ready (upon install/startup).

        :param timeout: Timeout in seconds.

        :returns: Tuple of parsed versions (multipass, multipassd).  multipassd
            may be None if Multipass is not ready and the timeout limit is reached.
        """
        if timeout is not None:
            deadline: Optional[float] = time.time() + timeout
        else:
            deadline = None

        while True:
            multipass_version, multipassd_version = self.version()

            if multipassd_version is not None:
                return (multipass_version, multipassd_version)

            if deadline is not None and time.time() >= deadline:
                break

            time.sleep(RETRY_WAIT)

        raise MultipassError(
            brief="Timed out waiting for Multipass to become ready.",
        )

    def version(self) -> Tuple[str, Optional[str]]:
        """Get multipass and multipassd versions.

        :returns: Tuple of parsed versions (multipass, multipassd).  multipassd
                  may be None if Multipass is not yet ready.
        """
        try:
            proc = self._run(["version"])
        except subprocess.CalledProcessError as error:
            raise MultipassError(
                brief="Failed to check version.",
                details=errors.details_from_called_process_error(error),
            ) from error

        try:
            output = proc.stdout.decode(encoding=locale.getpreferredencoding())
        except UnicodeDecodeError as error:
            raise MultipassError(
                brief="Failed to check version.",
                details=f"Failed to decode output: {proc.stdout!r}",
            ) from error

        # Expected multipass version output should look like:
        # * Scenario 1: multipassd not yet ready
        #
        #   multipass: 1.5.0
        #
        # * Scenario 2: typical Linux
        #
        #   multipass: 1.5.0
        #   multipassd: 1.5.0
        #
        # * Scenario 3: typical Mac
        #
        #   multipass: 1.5.0+mac
        #   multipassd: 1.5.0+mac
        #
        # * Scenario 4: typical Windows
        #
        #   multipass: 1.5.0+win
        #   multipassd: 1.5.0+win
        #
        # * Scenario 5: outdated Windows version with notice message
        #   See: https://github.com/canonical/multipass/issues/2020
        #
        #   multipass: 1.5.0+win
        #   multipassd: 1.5.0+win
        #
        #   SOME NOTICE INFORMATION....
        #
        # After stripping and splitting:
        #    - ['multipass', '1.5.0'] # noqa: ERA001
        #    - ['multipass', '1.5.0', 'multipassd', '1.5.0'] # noqa: ERA001
        #    - ['multipass', '1.5.0+mac', 'multipassd', '1.5.0+mac'] # noqa: ERA001
        #    - ['multipass', '1.5.0+win', 'multipassd', '1.5.0+win'] # noqa: ERA001
        #    - ['multipass', '1.5.0+win', 'multipassd', '1.5.0+win', ...] # noqa: ERA001
        output_split = output.strip().split()
        if len(output_split) < 2 or output_split[0] != "multipass":
            raise MultipassError(
                brief=f"Unable to parse version output: {proc.stdout!r}",
            )

        multipass_version = output_split[1].split("+")[0]

        if len(output_split) >= 4 and output_split[2] == "multipassd":
            multipassd_version = output_split[3].split("+")[0]
        else:
            multipassd_version = None

        return (multipass_version, multipassd_version)



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Multipass provider support package."""

from ._launch import launch
from ._ready import ensure_multipass_is_ready
from .errors import MultipassError, MultipassInstallationError
from .installer import install, is_installed
from .multipass import Multipass
from .multipass_instance import MultipassInstance
from .multipass_provider import MultipassProvider

__all__ = [
    "Multipass",
    "MultipassInstance",
    "MultipassError",
    "MultipassInstallationError",
    "MultipassProvider",
    "install",
    "is_installed",
    "ensure_multipass_is_ready",
    "launch",
]



#
# Copyright 2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Tests for craft_providers helper utilities."""
import sys
import time
from unittest import mock

import pytest
from craft_providers.util import retry

pytestmark = [
    # These tests can be flaky on Windows, at least in GitHub CI. The flakiness
    # comes from timing issues on Windows.
    pytest.mark.flaky(condition=sys.platform == "win32", reruns=3, reruns_delay=1),
]


@pytest.fixture(params=range(1, 6), ids=[f"timeout_{i}s" for i in range(1, 6)])
def timeout(request):
    return request.param


@pytest.fixture(params=range(1, 6), ids=[f"wait_{i}0ms" for i in range(1, 6)])
def retry_wait(request):
    """A parametrized fixture of times to wait before retrying, in microseconds"""
    return request.param * 0.01


def test_retry_until_timeout_success(
    failure_count, retry_wait, timeout, mock_instant_sleep
):
    failures = [Exception()] * failure_count
    mock_function = mock.Mock(side_effect=[*failures, None])

    retry.retry_until_timeout(timeout, retry_wait, mock_function)

    assert len(mock_function.mock_calls) == failure_count + 1
    assert len(mock_instant_sleep.sleep.mock_calls) >= failure_count


@pytest.mark.parametrize("retry_multiplier", [0.5, 0.8])
@pytest.mark.usefixtures("instant_sleep")
def test_retry_until_timeout_success_longish_retry(timeout, retry_multiplier):
    """retry_wait is more than half the timeout we always run twice."""
    retry_wait = timeout * retry_multiplier
    mock_function = mock.Mock(side_effect=[Exception(), None])

    retry.retry_until_timeout(timeout, retry_wait, mock_function)

    assert len(mock_function.mock_calls) == 2


@pytest.mark.parametrize("retry_multiplier", [1.0, 1000.0])
def test_retry_until_timeout_success_long_retry(monkeypatch, timeout, retry_multiplier):
    """retry_wait > timeout, we never loop."""
    retry_wait = retry_multiplier * timeout
    mock_function = mock.Mock(return_value=None)
    mock_monotonic = mock.Mock(wraps=time.monotonic)
    monkeypatch.setattr("time.monotonic", mock_monotonic)

    retry.retry_until_timeout(timeout, retry_wait, mock_function)

    mock_function.assert_called_once_with(retry_wait)
    assert mock_monotonic.mock_calls == [mock.call(), mock.call()]


@pytest.mark.parametrize("error_cls", [TimeoutError, Exception, ValueError])
def test_retry_until_timeout_times_out(retry_wait, timeout, instant_sleep, error_cls):
    mock_function = mock.Mock(side_effect=Exception())

    with pytest.raises(error_cls):
        retry.retry_until_timeout(timeout, retry_wait, mock_function, error=error_cls())

    mock_function.assert_called()


@pytest.mark.parametrize("error_cls", [TimeoutError, Exception, ValueError])
@pytest.mark.parametrize("retry_multiplier", [1.0, 1.1, 2.0])
def test_retry_until_timeout_times_out_long_retry(
    timeout, mock_instant_sleep, error_cls, retry_multiplier
):
    """Here the retry time is always at least as long as the timeout."""
    mock_function = mock.Mock(side_effect=Exception())

    with pytest.raises(error_cls):
        retry.retry_until_timeout(
            timeout, timeout * retry_multiplier, mock_function, error=error_cls()
        )

    mock_function.assert_called_once()
    mock_instant_sleep.sleep.assert_not_called()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from unittest.mock import call

import pytest
from craft_providers import lxd
from craft_providers.bases import ubuntu
from craft_providers.errors import BaseConfigurationError
from craft_providers.lxd import remotes


@pytest.fixture()
def mock_lxc(mocker):
    return mocker.patch("craft_providers.lxd.launcher.LXC", spec=lxd.LXC)


@pytest.fixture()
def fake_remote_image(mocker):
    return remotes.RemoteImage(
        image_name="test-image-name",
        remote_name="test-remote-name",
        remote_address="test-remote-address",
        remote_protocol=remotes.ProtocolType.LXD,
    )


@pytest.fixture()
def mock_remote_image(mocker):
    return mocker.patch(
        "craft_providers.lxd.remotes.RemoteImage", spec=remotes.RemoteImage
    )


@pytest.fixture()
def mock_get_remote_image(mocker, mock_remote_image):
    return mocker.patch(
        "craft_providers.lxd.remotes.get_remote_image", return_value=mock_remote_image
    )


@pytest.mark.parametrize(
    ("remote_name", "remote_address", "is_stable"),
    [
        (
            remotes.BUILDD_RELEASES_REMOTE_NAME,
            remotes.BUILDD_RELEASES_REMOTE_ADDRESS,
            True,
        ),
        (remotes.BUILDD_DAILY_REMOTE_NAME, remotes.BUILDD_DAILY_REMOTE_ADDRESS, False),
        (remotes.DAILY_REMOTE_NAME, remotes.DAILY_REMOTE_ADDRESS, False),
        ("other-remote-name", "other-remote-address", False),
    ],
)
def test_remote_image_is_stable_true(remote_name, remote_address, is_stable):
    """Verify `is_stable` is only true for images from the BUILDD_RELEASES remote."""
    test_remote = remotes.RemoteImage(
        image_name="test-image-name",
        remote_name=remote_name,
        remote_address=remote_address,
        remote_protocol=remotes.ProtocolType.LXD,
    )

    assert test_remote.is_stable == is_stable


def test_add_remote_new(fake_remote_image, mock_lxc, logs):
    """Verify remote is added if it does not already exist."""
    fake_remote_image.add_remote(mock_lxc)

    assert mock_lxc.mock_calls == [
        call.remote_list(),
        call.remote_list().__contains__("test-remote-name"),
        call.remote_add(
            remote="test-remote-name",
            addr="test-remote-address",
            protocol=remotes.ProtocolType.LXD.value,
        ),
    ]
    assert "Remote 'test-remote-name' was successfully added." in logs.debug


def test_add_remote_existing(fake_remote_image, mock_lxc, logs):
    """Verify existing remote is not added again."""
    mock_lxc.remote_list.return_value = {"test-remote-name": {}}

    fake_remote_image.add_remote(mock_lxc)

    assert mock_lxc.mock_calls == [call.remote_list()]
    assert "Remote 'test-remote-name' already exists." in logs.debug


def test_add_remote_race_condition(fake_remote_image, mock_lxc, logs):
    """Race condition when adding the remote, it was created by other process."""
    # the first time the list will not show the remote, then `remote_add` will fail
    # because it was just created from other process, and exactly because of that
    # it will show in the list the second time is asked
    mock_lxc.remote_list.side_effect = [{}, {"test-remote-name": {}}]
    mock_lxc.remote_add.side_effect = ValueError()

    fake_remote_image.add_remote(mock_lxc)

    assert mock_lxc.mock_calls == [
        call.remote_list(),
        call.remote_add(
            remote="test-remote-name",
            addr="test-remote-address",
            protocol=remotes.ProtocolType.LXD.value,
        ),
        call.remote_list(),
    ]
    assert (
        "Remote 'test-remote-name' is present on second check, ignoring exception "
        "ValueError()." in logs.debug
    )


def test_add_remote_race_condition_error(fake_remote_image, mock_lxc, logs):
    """Race condition code triggered but it was actually an error."""
    mock_lxc.remote_list.return_value = {}
    mock_lxc.remote_add.side_effect = ValueError()

    with pytest.raises(ValueError):
        fake_remote_image.add_remote(mock_lxc)


@pytest.mark.parametrize(
    ("provider_base_alias", "image_name"),
    [
        (ubuntu.BuilddBaseAlias.BIONIC, "core18"),
        (ubuntu.BuilddBaseAlias.FOCAL, "core20"),
        (ubuntu.BuilddBaseAlias.JAMMY, "core22"),
        (ubuntu.BuilddBaseAlias.LUNAR, "lunar"),
        (ubuntu.BuilddBaseAlias.DEVEL, "devel"),
    ],
)
def test_get_image_remote(provider_base_alias, image_name):
    """Verify `get_remote_image()` returns a RemoteImage."""
    base = ubuntu.BuilddBase(alias=provider_base_alias)
    remote_image = lxd.remotes.get_remote_image(base)

    assert remote_image.image_name == image_name


@pytest.mark.parametrize(
    ("provider_base_alias", "image_name"),
    [
        (ubuntu.BuilddBaseAlias.BIONIC, "core18"),
        (ubuntu.BuilddBaseAlias.FOCAL, "core20"),
        (ubuntu.BuilddBaseAlias.JAMMY, "core22"),
        (ubuntu.BuilddBaseAlias.LUNAR, "lunar"),
        (ubuntu.BuilddBaseAlias.DEVEL, "devel"),
    ],
)
def test_get_image_remote_deprecated(provider_base_alias, image_name):
    """Verify `get_remote_image()` returns a RemoteImage.
    temporary backward compatibility before 2.0
    """
    remote_image = lxd.remotes.get_remote_image(str(provider_base_alias.value))

    assert remote_image.image_name == image_name


def test_get_image_remote_xenial_error():
    """Raise an error when retrieving a xenial image.

    The remote image for Xenial has not been chosen for craft-providers + LXD, so an
    error is raised.
    """
    base = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.XENIAL)
    with pytest.raises(lxd.LXDError) as raised:
        lxd.remotes.get_remote_image(base)

    assert "could not find a lxd remote image for the provider base " in str(
        raised.value
    )


def test_get_image_remote_error():
    """Raise an error for an unknown provider base."""
    base = ubuntu.BuilddBase(alias=-1)  # type: ignore
    with pytest.raises(lxd.LXDError) as raised:
        lxd.remotes.get_remote_image(base)

    assert "could not find a lxd remote image for the provider base " in str(
        raised.value
    )


def test_get_image_remote_deprecated_error():
    """Raise an error for an unknown provider base."""
    with pytest.raises(BaseConfigurationError) as raised:
        lxd.remotes.get_remote_image("8.04")

    assert (
        str(raised.value)
        == "Base alias not found for BaseName(name='ubuntu', version='8.04')"
    )


def test_configure_buildd_image_remote(
    mock_lxc, logs, mock_get_remote_image, mock_remote_image
):
    """Verify deprecated `configure_buildd_image_remote()` call."""

    with pytest.warns(DeprecationWarning) as warning:
        name = lxd.remotes.configure_buildd_image_remote(lxc=mock_lxc)

    assert str(warning[0].message) == (
        "configure_buildd_image_remote() is deprecated. "
        "Use get_remote_image() and RemoteImage.add_remote()."
    )
    mock_get_remote_image.assert_called_once()
    mock_remote_image.add_remote.assert_called_once_with(mock_lxc)
    assert name == remotes.BUILDD_RELEASES_REMOTE_NAME



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import io
import json
import os
import pathlib
import subprocess
import sys
from datetime import datetime, timedelta
from threading import Thread

import pytest
from craft_providers import lxd
from craft_providers.bases import ubuntu
from craft_providers.errors import BaseCompatibilityError
from craft_providers.lxd import project as lxd_project
from freezegun import freeze_time

from . import conftest


@pytest.fixture(scope="session")
def get_base_instance():
    def _base_instance(
        image_name: str = "22.04",
        image_remote: str = "ubuntu",
        compatibility_tag: str = "buildd-base-v2",
        project: str = "default",
    ):
        """Get the base instance."""
        base_instance_name = lxd.launcher._formulate_base_instance_name(
            image_name=image_name,
            image_remote=image_remote,
            compatibility_tag=compatibility_tag,
        )
        instance = lxd.LXDInstance(name=base_instance_name, project=project)
        return instance

    return _base_instance


@pytest.fixture()
def base_configuration():
    """Returns a simple base configuration."""
    return ubuntu.BuilddBase(
        alias=ubuntu.BuilddBaseAlias.JAMMY,
        hostname="test-hostname",
    )


@pytest.fixture()
def core22_instance(instance_name):
    """Yields a minimally setup core22 instance.

    The yielded instance will be launched, started, and marked as setup, even though
    most of the setup is skipped to speed up test execution.

    Delete instance on fixture teardown.
    """
    with conftest.tmp_instance(
        name=instance_name,
        image="22.04",
        image_remote="ubuntu",
        project="default",
    ):
        instance = lxd.LXDInstance(name=instance_name)

        # mark instance as setup in the config file
        instance.push_file_io(
            destination=ubuntu.BuilddBase._instance_config_path,
            content=io.BytesIO(
                f"compatibility_tag: {ubuntu.BuilddBase.compatibility_tag}"
                "\nsetup: true\n".encode()
            ),
            file_mode="0644",
        )

        yield instance

        if instance.exists():
            instance.delete()


@pytest.fixture()
def get_instance_and_base_instance(
    base_configuration, get_base_instance, instance_name
):
    """Create and return an instance and base instance as a tuple.

    Delete instances on fixture teardown.
    """
    base_instance = get_base_instance()

    # launch an instance from an image and create a base instance
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_base_instance=True,
    )
    try:
        # the instance and base instance should both exist
        assert instance.exists()
        assert base_instance.exists()

        # only the instance should be running
        assert instance.is_running()
        assert not base_instance.is_running()
        yield (instance, base_instance)
    finally:
        if instance.exists():
            instance.delete()
        if base_instance.exists():
            base_instance.delete()


def test_launch_and_run(instance_name):
    """Launch an instance from the `ubuntu` remote and run a command in the instance."""
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name=ubuntu.BuilddBaseAlias.JAMMY.value,
        image_remote="ubuntu",
    )

    try:
        assert isinstance(instance, lxd.LXDInstance)
        assert instance.exists()
        assert instance.is_running()

        proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)

        assert proc.stdout == b"hi\n"
    finally:
        instance.delete()


@pytest.mark.skipif(sys.platform != "linux", reason=f"unsupported on {sys.platform}")
def test_timezone(instance_name):
    """Verify instance timezone matches the host timezone."""
    base_configuration = ubuntu.BuilddBase(alias=ubuntu.BuilddBaseAlias.JAMMY)

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name=ubuntu.BuilddBaseAlias.JAMMY.value,
        image_remote="ubuntu",
    )

    try:
        host_timezone = subprocess.run(
            ["timedatectl", "show", "-p", "Timezone", "--value"],
            capture_output=True,
            check=True,
            text=True,
        ).stdout.strip()

        instance_timezone = instance.execute_run(
            ["printenv", "TZ"], capture_output=True, check=True, text=True
        ).stdout.strip()

        assert host_timezone == instance_timezone

    finally:
        instance.delete()


def test_launch_use_snapshots_deprecated(
    base_configuration, get_base_instance, instance_name
):
    """Launch an instance with the deprecated parameter `use_snapshots`."""
    base_instance = get_base_instance()

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_snapshots=True,
    )
    try:
        # verify the instance and base instance exist
        assert instance.exists()
        assert base_instance.exists()
    finally:
        if instance.exists():
            instance.delete()
        if base_instance.exists():
            base_instance.delete()


def test_launch_use_base_instance(
    base_configuration, get_instance_and_base_instance, instance_name
):
    """Launch an instance using base instances.

    First, launch an instance from an image and create a base instance.
    Then launch an instance from the base instance.
    Then launch an instance when the instance exists.

    The parameter `use_base_instance` and the deprecated parameter `use_snapshots`
    should both result in the same behavior.
    """
    instance, base_instance = get_instance_and_base_instance

    # fingerprint the base instance
    base_instance.start()
    base_instance.execute_run(["touch", "/base-instance"])
    base_instance.stop()

    # delete the instance so a new instance is created from the base instance
    instance.delete()
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_base_instance=True,
    )

    assert instance.exists()
    assert instance.is_running()

    # confirm instance was created from the base instance by checking fingerprint
    instance.execute_run(["stat", "/base-instance"], check=True)

    # fingerprint the instance
    instance.execute_run(["touch", "/instance"])

    # relaunch the existing instance
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_base_instance=True,
    )

    # confirm the same instance was launched with both fingerprints
    instance.execute_run(["stat", "/base-instance"], check=True)
    instance.execute_run(["stat", "/instance"], check=True)

    assert instance.exists()
    assert instance.is_running()

    # collect the hostname of the instance
    instance_hostname = (
        instance.execute_run(["hostname"], check=True, capture_output=True)
        .stdout.decode()
        .rstrip("\n")
    )

    assert instance_hostname == instance.instance_name


def test_launch_create_base_instance_with_correct_image_description(
    base_configuration, get_base_instance, instance_name
):
    """Create a base instance and check the image description"""
    base_instance = get_base_instance()

    # launch an instance from an image and create a base instance
    lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_base_instance=True,
    )

    lxc_result_json = subprocess.check_output(
        [
            "lxc",
            "--project",
            base_instance.project,
            "--format",
            "json",
            "list",
            base_instance.instance_name,
        ]
    ).decode()

    lxc_result = json.loads(lxc_result_json)

    assert (
        lxc_result[0]["expanded_config"]["image.description"]
        == "base-instance-buildd-base-v2-ubuntu-22.04"
    )


@freeze_time(datetime.now() + timedelta(days=91))
def test_launch_use_base_instance_expired(
    base_configuration, get_instance_and_base_instance, instance_name
):
    """Launch an instance using an expired base instance.

    First, launch an instance from an image and create a base instance.
    Then launch an instance from the expired base instance, which will trigger the
    creation of a new instance and base instance.

    The LXD instance is created via subprocess, so the creation date the instance is
    out of freezegun's scope and can't be modified.
    """
    instance, base_instance = get_instance_and_base_instance

    # fingerprint the expired base instance
    base_instance.start()
    base_instance.execute_run(["touch", "/base-instance"])
    base_instance.stop()

    # delete the instance so a new instance is created from the base instance
    instance.delete()
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_base_instance=True,
        expiration=timedelta(days=90),
    )

    assert instance.exists()
    assert instance.is_running()

    # confirm instance does not have the expired base instance's fingerprint
    proc = instance.execute_run(
        ["stat", "/base-instance"], capture_output=True, text=True
    )
    assert proc.returncode == 1
    assert "'/base-instance': No such file or directory" in proc.stderr

    # confirm new base instance does not have the expired base instance's fingerprint
    base_instance.start()
    proc = base_instance.execute_run(
        ["stat", "/base-instance"], capture_output=True, text=True
    )
    assert proc.returncode == 1
    assert "'/base-instance': No such file or directory" in proc.stderr


def test_launch_create_project(base_configuration, instance_name, project_name):
    """Create a project if it does not exist and `auto_create_project` is true."""
    lxc = lxd.LXC()

    assert project_name not in lxc.project_list()

    try:
        instance = lxd.launch(
            name=instance_name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
            auto_create_project=True,
            project=project_name,
            remote="local",
        )

        assert instance.exists()
        assert project_name in lxc.project_list()
    finally:
        lxd_project.purge(lxc=lxc, project=project_name)


def test_launch_with_project_and_use_base_instance(
    base_configuration, get_base_instance, instance_name, session_project
):
    """With a LXD project specified, launch an instance and use base instances."""
    base_instance = get_base_instance(project=session_project)

    # launch an instance from an image and create a base instance
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        use_base_instance=True,
        project=session_project,
        remote="local",
    )

    try:
        # the instance and base instance should both exist
        assert instance.exists()
        assert base_instance.exists()

        # only the instance should be running
        assert instance.is_running()
        assert not base_instance.is_running()

        # delete the instance so a new instance is created from the base instance
        instance.delete()
        instance = lxd.launch(
            name=instance_name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
            use_base_instance=True,
            project=session_project,
            remote="local",
        )

        assert instance.exists()
        assert instance.is_running()

        # relaunch the existing instance
        instance = lxd.launch(
            name=instance_name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
            use_base_instance=True,
            project=session_project,
            remote="local",
        )

        assert instance.exists()
        assert instance.is_running()
    finally:
        if instance.exists():
            instance.delete()
        if base_instance.exists():
            base_instance.delete()


def test_launch_with_project_and_use_base_instance_parallel(
    base_configuration, get_base_instance, instance_name, session_project
):
    """Launch 5 instances at same time and use the same base instances."""
    base_instance = get_base_instance(project=session_project)

    instances = []

    class InstanceThread(Thread):
        def __init__(self, instance_name):
            super().__init__()
            self.instance_name = instance_name
            self.instance = None

        def run(self):
            self.instance = lxd.launch(
                name=self.instance_name,
                base_configuration=base_configuration,
                image_name="22.04",
                image_remote="ubuntu",
                use_base_instance=True,
                project=session_project,
                remote="local",
            )

    # launch 5 instances from an image and create a base instance
    for i in range(0, 5):
        instance_thread = InstanceThread(f"{instance_name}-{i}")
        instances.append(instance_thread)

    for instance_thread in instances:
        instance_thread.start()

    for instance_thread in instances:
        instance_thread.join()

    try:
        assert base_instance.exists()
        assert not base_instance.is_running()

        for instance_thread in instances:
            assert instance_thread.instance is not None
            assert instance_thread.instance.exists()
            assert instance_thread.instance.is_running()

    finally:
        for instance_thread in instances:
            if (
                instance_thread.instance is not None
                and instance_thread.instance.exists()
            ):
                instance_thread.instance.delete()

        if base_instance.exists():
            base_instance.delete()


def test_launch_ephemeral(base_configuration, instance_name):
    """Launch an ephemeral instance and verify it is deleted after it is stopped."""

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        ephemeral=True,
    )

    try:
        # lxd will delete the instance when it is stopped
        instance.stop()

        assert not instance.exists()
    finally:
        if instance.exists():
            instance.delete()


def test_launch_ephemeral_existing(base_configuration, instance_name):
    """If an ephemeral instance already exists, delete it and create a new instance."""

    # create a non-ephemeral instance
    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        ephemeral=False,
    )

    try:
        assert instance.exists()
        assert instance.is_running()

        # relaunching as an ephemeral instance will delete the existing instance
        instance = lxd.launch(
            name=instance_name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
            ephemeral=True,
        )

        assert instance.exists()

        # lxd will delete the instance when it is stopped
        instance.stop()

        assert not instance.exists()
    finally:
        if instance.exists():
            instance.delete()


def test_launch_map_user_uid_true(base_configuration, instance_name, tmp_path):
    """Enable and map the the UID of the test account."""
    tmp_path.chmod(0o755)

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        map_user_uid=True,
        uid=os.stat(tmp_path).st_uid,
    )

    try:
        instance.mount(host_source=tmp_path, target=pathlib.PurePosixPath("/mnt"))

        # If user ID mappings are enabled, we will be able to write.
        instance.execute_run(["touch", "/mnt/foo"], capture_output=True, check=True)
    finally:
        if instance.exists():
            instance.delete()


def test_launch_map_user_uid_true_no_uid(base_configuration, instance_name, tmp_path):
    """Enable UID mapping without specifying a UID."""
    tmp_path.chmod(0o755)

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        map_user_uid=True,
    )

    try:
        instance.mount(host_source=tmp_path, target=pathlib.PurePosixPath("/mnt"))

        # If user ID mappings are enabled, we will be able to write.
        instance.execute_run(["touch", "/mnt/foo"], capture_output=True, check=True)
    finally:
        if instance.exists():
            instance.delete()


def test_launch_map_user_uid_false(base_configuration, instance_name, tmp_path):
    """If UID mapping is not enabled, access to a mounted directory will be denied."""
    tmp_path.chmod(0o755)

    instance = lxd.launch(
        name=instance_name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        map_user_uid=False,
    )

    try:
        instance.mount(host_source=tmp_path, target=pathlib.PurePosixPath("/mnt"))

        # If user ID mappings are not enabled, we won't be able to write.
        with pytest.raises(subprocess.CalledProcessError):
            instance.execute_run(["touch", "/mnt/foo"], capture_output=True, check=True)
    finally:
        if instance.exists():
            instance.delete()


def test_launch_existing_instance(base_configuration, core22_instance):
    """Launch an existing instance and run a command."""
    instance = lxd.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
    )

    assert instance.exists()
    assert instance.is_running()

    proc = instance.execute_run(["echo", "hi"], check=True, stdout=subprocess.PIPE)

    assert proc.stdout == b"hi\n"


def test_launch_os_incompatible_without_auto_clean(base_configuration, core22_instance):
    """Raise an error if the OS is incompatible and auto_clean is False."""
    core22_instance.push_file_io(
        destination=pathlib.PurePosixPath("/etc/os-release"),
        content=io.BytesIO(b"NAME=Fedora\nVERSION_ID=32\n"),
        file_mode="0644",
    )

    # will raise compatibility error when auto_clean is false
    with pytest.raises(BaseCompatibilityError) as exc_info:
        lxd.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
        )

    assert (
        exc_info.value.brief
        == "Incompatible base detected: Expected OS 'Ubuntu', found 'Fedora'."
    )


def test_launch_os_incompatible_with_auto_clean(base_configuration, core22_instance):
    """Clean the instance if the OS is incompatible and auto_clean is True."""
    core22_instance.push_file_io(
        destination=pathlib.PurePosixPath("/etc/os-release"),
        content=io.BytesIO(b"NAME=Fedora\nVERSION_ID=32\n"),
        file_mode="0644",
    )

    # when auto_clean is true, the instance will be deleted and recreated
    lxd.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        auto_clean=True,
    )

    assert core22_instance.exists()
    assert core22_instance.is_running()


def test_launch_instance_config_incompatible_without_auto_clean(
    base_configuration, core22_instance
):
    """Raise an error if the config file is incompatible and auto_clean is False."""
    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: invalid\nsetup: true\n"),
        file_mode="0644",
    )

    # will raise compatibility error when auto_clean is false
    with pytest.raises(BaseCompatibilityError) as exc_info:
        lxd.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
        )

    assert exc_info.value.brief == (
        "Incompatible base detected:"
        " Expected image compatibility tag 'buildd-base-v2', found 'invalid'."
    )


def test_launch_instance_config_incompatible_with_auto_clean(
    base_configuration, core22_instance
):
    """Clean the instance if the config is incompatible and auto_clean is True."""
    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: invalid\nsetup: true\n"),
        file_mode="0644",
    )

    # when auto_clean is true, the instance will be deleted and recreated
    lxd.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        auto_clean=True,
    )

    assert core22_instance.exists()
    assert core22_instance.is_running()


def test_launch_instance_not_setup_without_auto_clean(
    base_configuration, core22_instance
):
    """Raise an error if an existing instance is not setup and auto_clean is False."""
    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: buildd-base-v2\nsetup: false\n"),
        file_mode="0644",
    )

    # will raise a compatibility error
    with pytest.raises(BaseCompatibilityError) as exc_info:
        lxd.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
            auto_clean=False,
        )

    assert exc_info.value == BaseCompatibilityError("instance is marked as not setup")


def test_launch_instance_not_setup_with_auto_clean(base_configuration, core22_instance):
    """Clean the instance if it is not setup and auto_clean is True."""
    core22_instance.push_file_io(
        destination=base_configuration._instance_config_path,
        content=io.BytesIO(b"compatibility_tag: buildd-base-v2\nsetup: false\n"),
        file_mode="0644",
    )

    # when auto_clean is true, the instance will be deleted and recreated
    lxd.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        auto_clean=True,
    )

    assert core22_instance.exists()
    assert core22_instance.is_running()


def test_launch_instance_id_map_incompatible_without_auto_clean(
    base_configuration, core22_instance
):
    """Raise an error if the id map is incompatible and auto_clean is False."""
    lxc = lxd.LXC()
    lxc.config_set(
        instance_name=core22_instance.instance_name,
        key="raw.idmap",
        value="",  # equivalent to `lxc config unset`, which is not yet implemented
    )

    # will raise compatibility error when auto_clean is false
    with pytest.raises(BaseCompatibilityError) as exc_info:
        lxd.launch(
            name=core22_instance.name,
            base_configuration=base_configuration,
            image_name="22.04",
            image_remote="ubuntu",
            map_user_uid=True,
        )

    assert exc_info.value.brief == (
        "Incompatible base detected: "
        "the instance's id map ('raw.idmap') is not configured as expected."
    )


def test_launch_instance_id_map_incompatible_with_auto_clean(
    base_configuration, core22_instance
):
    """Clean the instance if the id map is incompatible and auto_clean is True."""
    lxc = lxd.LXC()
    lxc.config_set(
        instance_name=core22_instance.instance_name,
        key="raw.idmap",
        value="",  # equivalent to `lxc config unset`, which is not yet implemented
    )

    # when auto_clean is true, the instance will be deleted and recreated
    lxd.launch(
        name=core22_instance.name,
        base_configuration=base_configuration,
        image_name="22.04",
        image_remote="ubuntu",
        auto_clean=True,
    )

    assert core22_instance.exists()
    assert core22_instance.is_running()



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Tests for snap installer."""

import logging
import re
from textwrap import dedent

from craft_providers.actions import snap_installer


def test_inject_from_host(core22_lxd_instance, installed_snap, caplog):
    """Verify a snap can be injected from the host."""
    core22_lxd_instance.execute_run(
        ["test", "!", "-d", "/snap/hello-world"], check=True
    )

    with installed_snap("hello-world"):
        snap_installer.inject_from_host(
            executor=core22_lxd_instance, snap_name="hello-world", classic=False
        )

    core22_lxd_instance.execute_run(["test", "-d", "/snap/hello-world"], check=True)

    config = core22_lxd_instance.execute_run(
        ["cat", "/etc/craft-instance.conf"], check=True, capture_output=True
    ).stdout.decode()

    # verify instance config was properly updated
    assert (
        re.fullmatch(
            re.compile(
                dedent(
                    """\
                    compatibility_tag: buildd-base-v\\d+
                    setup: true
                    snaps:
                      hello-world:
                        revision: '\\d+'
                        source: host
                    """
                )
            ),
            config,
        )
        is not None
    )
    assert caplog.records == []


def test_inject_from_host_dangerous(
    core22_lxd_instance, dangerously_installed_snap, caplog
):
    """Verify a dangerously installed snap can be injected from the host."""
    core22_lxd_instance.execute_run(
        ["test", "!", "-d", "/snap/hello-world"], check=True
    )

    with dangerously_installed_snap("hello-world"):
        snap_installer.inject_from_host(
            executor=core22_lxd_instance, snap_name="hello-world", classic=False
        )

    core22_lxd_instance.execute_run(["test", "-d", "/snap/hello-world"], check=True)

    config = core22_lxd_instance.execute_run(
        ["cat", "/etc/craft-instance.conf"], check=True, capture_output=True
    ).stdout.decode()

    # verify instance config was properly updated
    assert (
        re.fullmatch(
            re.compile(
                dedent(
                    """\
                    compatibility_tag: buildd-base-v\\d+
                    setup: true
                    snaps:
                      hello-world:
                        revision: x\\d+
                        source: host
                    """
                )
            ),
            config,
        )
        is not None
    )
    assert caplog.records == []


def test_inject_from_host_using_pack_fallback(
    core22_lxd_instance, empty_test_snap, caplog
):
    """Verify a snap is packed if the local download fails."""
    caplog.set_level(logging.DEBUG)

    snap_installer.inject_from_host(
        executor=core22_lxd_instance,
        snap_name=empty_test_snap,
        classic=False,
    )

    core22_lxd_instance.execute_run(
        ["test", "-d", f"/snap/{empty_test_snap}"], check=True
    )

    log_messages = [r.message for r in caplog.records]
    expected = (
        "Failed to fetch snap from snapd, falling back to `snap pack` to recreate"
    )
    assert expected in log_messages


def test_install_from_store_strict(core22_lxd_instance, installed_snap, caplog):
    """Verify a strictly confined snap from the store can be installed."""
    core22_lxd_instance.execute_run(
        ["test", "!", "-d", "/snap/hello-world"], check=True
    )

    snap_installer.install_from_store(
        executor=core22_lxd_instance,
        snap_name="hello-world",
        channel="latest/stable",
        classic=False,
    )

    core22_lxd_instance.execute_run(["test", "-f", "/snap/bin/hello-world"], check=True)

    config = core22_lxd_instance.execute_run(
        ["cat", "/etc/craft-instance.conf"], check=True, capture_output=True
    ).stdout.decode()

    # verify instance config was properly updated
    assert (
        re.fullmatch(
            re.compile(
                dedent(
                    """\
                    compatibility_tag: buildd-base-v\\d+
                    setup: true
                    snaps:
                      hello-world:
                        revision: '\\d+'
                        source: store
                    """
                )
            ),
            config,
        )
        is not None
    )
    assert caplog.records == []


def test_install_from_store_classic(core22_lxd_instance, installed_snap, caplog):
    """Verify a classicly confined snap from the store can be installed."""
    core22_lxd_instance.execute_run(["test", "!", "-d", "/snap/charmcraft"], check=True)

    snap_installer.install_from_store(
        executor=core22_lxd_instance,
        snap_name="charmcraft",
        channel="latest/stable",
        classic=True,
    )

    core22_lxd_instance.execute_run(["test", "-f", "/snap/bin/charmcraft"], check=True)

    config = core22_lxd_instance.execute_run(
        ["cat", "/etc/craft-instance.conf"], check=True, capture_output=True
    ).stdout.decode()

    # verify instance config was properly updated
    assert (
        re.fullmatch(
            re.compile(
                dedent(
                    """\
                    compatibility_tag: buildd-base-v\\d+
                    setup: true
                    snaps:
                      charmcraft:
                        revision: '\\d+'
                        source: store
                    """
                )
            ),
            config,
        )
        is not None
    )
    assert caplog.records == []


def test_install_from_store_channel(core22_lxd_instance, installed_snap, caplog):
    """Verify a channel can be specified when installing from the store"""
    core22_lxd_instance.execute_run(["test", "!", "-d", "/snap/go"], check=True)

    snap_installer.install_from_store(
        executor=core22_lxd_instance,
        snap_name="go",
        channel="1.15/stable",
        classic=True,
    )

    proc = core22_lxd_instance.execute_run(
        ["/snap/bin/go", "version"], capture_output=True, check=True, text=True
    )

    assert "go1.15" in proc.stdout

    config = core22_lxd_instance.execute_run(
        ["cat", "/etc/craft-instance.conf"], check=True, capture_output=True
    ).stdout.decode()

    # verify instance config was properly updated
    assert (
        re.fullmatch(
            re.compile(
                dedent(
                    """\
                    compatibility_tag: buildd-base-v\\d+
                    setup: true
                    snaps:
                      go:
                        revision: '\\d+'
                        source: store
                    """
                )
            ),
            config,
        )
        is not None
    )
    assert caplog.records == []


def test_install_from_store_snap_name_suffix(
    core22_lxd_instance, installed_snap, caplog
):
    """Verify a snap can be installed from the store when the snap name has a suffix."""
    core22_lxd_instance.execute_run(
        ["test", "!", "-d", "/snap/hello-world"], check=True
    )

    snap_installer.install_from_store(
        executor=core22_lxd_instance,
        snap_name="hello-world_suffix",
        channel="latest/stable",
        classic=False,
    )

    core22_lxd_instance.execute_run(["test", "-f", "/snap/bin/hello-world"], check=True)

    config = core22_lxd_instance.execute_run(
        ["cat", "/etc/craft-instance.conf"], check=True, capture_output=True
    ).stdout.decode()

    # track the `hello-world` snap in the config file, not `hello-world_suffix`
    assert (
        re.fullmatch(
            re.compile(
                dedent(
                    """\
                    compatibility_tag: buildd-base-v\\d+
                    setup: true
                    snaps:
                      hello-world:
                        revision: '\\d+'
                        source: store
                    """
                )
            ),
            config,
        )
        is not None
    )
    assert caplog.records == []



#
# Copyright 2021-2023 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pathlib
from unittest import mock

import pytest
import yaml
from craft_providers import Executor
from craft_providers.errors import BaseConfigurationError, ProviderError
from craft_providers.instance_config import InstanceConfiguration
from pydantic import ValidationError


@pytest.fixture()
def default_config_data():
    return {
        "compatibility_tag": "tag-foo-v2",
        "setup": True,
        "snaps": {
            "charmcraft": {"revision": 834},
            "core22": {"revision": 147},
        },
    }


@pytest.fixture()
def mock_executor():
    executor_mock = mock.Mock(spec=Executor)
    return executor_mock


@pytest.fixture()
def config_fixture(fake_home_temporary_file):
    """Creates an instance config file containing data passed to the fixture."""

    def _config_fixture(**kwargs):
        fake_home_temporary_file.write_text(**kwargs)

        return fake_home_temporary_file

    return _config_fixture


def test_instance_config_defaults():
    """Verify default values for instance configuration objects."""
    config = InstanceConfiguration()

    assert config.setup is None
    assert config.compatibility_tag is None
    assert config.snaps is None


def test_save(mock_executor):
    config = InstanceConfiguration(compatibility_tag="tag-foo-v2")
    config_path = pathlib.PurePosixPath("/etc/crafty-crafty.conf")

    config.save(executor=mock_executor, config_path=config_path)

    assert mock_executor.mock_calls == [
        mock.call.push_file_io(
            destination=config_path, content=mock.ANY, file_mode="0644"
        )
    ]

    assert (
        mock_executor.mock_calls[0].kwargs["content"].read()
        == b"compatibility_tag: tag-foo-v2\n"
    )


def test_load_missing_config_returns_none(mock_executor):
    mock_executor.pull_file.side_effect = [FileNotFoundError]
    config_path = pathlib.PurePosixPath("/test/foo")
    config_instance = InstanceConfiguration.load(
        executor=mock_executor, config_path=config_path
    )

    assert config_instance is None


def test_load_empty_config_returns_none(mock_executor, config_fixture):
    config_fixture(data="")
    config_path = pathlib.PurePosixPath("/etc/crafty-crafty.conf")
    config_instance = InstanceConfiguration.load(
        executor=mock_executor, config_path=config_path
    )

    assert config_instance is None


def test_load_with_valid_config(mock_executor, config_fixture, default_config_data):
    config_file = config_fixture(data=yaml.dump(default_config_data))
    config_path = pathlib.PurePosixPath("/etc/crafty-crafty.conf")
    config_instance = InstanceConfiguration.load(
        executor=mock_executor, config_path=config_path
    )

    assert mock_executor.mock_calls == [
        mock.call.pull_file(
            source=config_path,
            destination=config_file,
        ),
    ]

    assert config_instance is not None
    assert dict(config_instance) == {
        "compatibility_tag": "tag-foo-v2",
        "setup": True,
        "snaps": {"charmcraft": {"revision": 834}, "core22": {"revision": 147}},
    }


def test_load_with_invalid_config_raises_error(mock_executor, config_fixture):
    config_fixture(data="invalid: data")

    config_path = pathlib.PurePosixPath("/etc/crafty-crafty.conf")

    with pytest.raises(ValidationError) as exc_info:
        InstanceConfiguration.load(executor=mock_executor, config_path=config_path)

    error = exc_info.value.errors()
    assert len(error) == 1
    assert error[0]["loc"] == ("invalid",)
    assert error[0]["type"] in ("value_error.extra", "extra_forbidden")


def test_load_failure_to_pull_file_raises_error(mock_executor):
    mock_executor.pull_file.side_effect = [ProviderError(brief="foo")]

    config_path = pathlib.PurePosixPath("/test/foo")

    with pytest.raises(BaseConfigurationError) as exc_info:
        InstanceConfiguration.load(executor=mock_executor, config_path=config_path)

    assert exc_info.value == BaseConfigurationError(
        brief=f"Failed to read instance config in environment at {config_path}",
    )


def test_update_single_value(default_config_data, mock_executor, mocker, tmpdir):
    """Test that a single value in a config is properly updated."""
    mocker.patch(
        "craft_providers.instance_config.InstanceConfiguration.load",
        return_value=InstanceConfiguration(**default_config_data),
    )

    updated_config = InstanceConfiguration.update(
        executor=mock_executor,
        data={"compatibility_tag": "updated-tag-value"},
        config_path=pathlib.Path(tmpdir / "crafty-crafty.conf"),
    )

    assert updated_config.compatibility_tag == "updated-tag-value"


def test_update_update_nested_values(
    default_config_data, mock_executor, mocker, tmpdir
):
    """Test updating a config by updating an existing nested value."""
    mocker.patch(
        "craft_providers.instance_config.InstanceConfiguration.load",
        return_value=InstanceConfiguration(**default_config_data),
    )

    updated_config = InstanceConfiguration.update(
        executor=mock_executor,
        data={
            "snaps": {
                "charmcraft": {"revision": 835},
            }
        },
        config_path=pathlib.Path(tmpdir / "crafty-crafty.conf"),
    )

    assert updated_config.snaps == {
        "charmcraft": {"revision": 835},
        "core22": {"revision": 147},
    }


def test_update_add_nested_values(default_config_data, mock_executor, mocker, tmpdir):
    """Test updating a config by adding a new nested value."""
    mocker.patch(
        "craft_providers.instance_config.InstanceConfiguration.load",
        return_value=InstanceConfiguration(**default_config_data),
    )

    updated_config = InstanceConfiguration.update(
        executor=mock_executor,
        data={
            "snaps": {
                "new-test-snap": {"revision": 1},
            }
        },
        config_path=pathlib.Path(tmpdir / "crafty-crafty.conf"),
    )

    assert updated_config.snaps == {
        "charmcraft": {"revision": 834},
        "core22": {"revision": 147},
        "new-test-snap": {"revision": 1},
    }



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import json
import pathlib
import subprocess
import textwrap
from unittest import mock

import pytest
import requests
import yaml
from craft_providers.actions import snap_installer
from craft_providers.actions.snap_installer import Snap
from craft_providers.errors import (
    BaseConfigurationError,
    ProviderError,
    details_from_called_process_error,
)
from craft_providers.instance_config import InstanceConfiguration
from logassert import Exact  # type: ignore


@pytest.fixture()
def mock_requests():
    """Mock requests_unixsocket."""
    with mock.patch(
        "craft_providers.actions.snap_installer.requests_unixsocket"
    ) as mock_requests:
        yield mock_requests


@pytest.fixture(params=["1"])
def config_fixture(fake_home_temporary_file, request):
    """Creates an instance config file in the pytest temp directory.

    Patches the temp_paths functions to point to the
    temporary directory containing the config.

    :param request: The revision of the target's snap. Default revision = 1.
    """
    config_content = textwrap.dedent(
        f"""\
        compatibility_tag: tag-foo-v2
        snaps:
          test-name:
            revision: '{request.param}'
            source: {snap_installer.SNAP_SRC_HOST}
        """
    )

    fake_home_temporary_file.write_text(config_content)


@pytest.fixture(params=[{"revision": "2", "id": "3", "publisher": {"id": "4"}}])
def mock_get_host_snap_info(request, mocker):
    """Mocks the get_host_snap_revision() function

    :param request: The revision of the host's snap. Default revision = 2.
    """
    mockee = "craft_providers.actions.snap_installer.get_host_snap_info"
    if isinstance(request.param, Exception):
        mocker.patch(mockee, side_effect=request.param)
    else:
        mocker.patch(mockee, return_value=request.param)


@pytest.fixture(params=["3"])
def mock_get_snap_revision_ensuring_source(request, mocker):
    """Mocks the _get_snap_revision_ensuring_source() function

    :param request: The revision of the snap. Default revision = 3.
    """
    mocker.patch(
        "craft_providers.actions.snap_installer._get_snap_revision_ensuring_source",
        return_value=request.param,
    )


@pytest.fixture(params=["4"])
def mock_get_target_snap_revision_from_snapd(request, mocker):
    """Mocks the _get_target_snap_revision_from_snapd() function

    :param request: The revision of the snap. Default revision = 4.
    """
    mocker.patch(
        "craft_providers.actions.snap_installer._get_target_snap_revision_from_snapd",
        return_value=request.param,
    )


def test_inject_from_host_classic(
    config_fixture,
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    logs,
    tmp_path,
):
    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "ack",
            "/tmp/test-name.assert",
        ]
    )
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
            "--classic",
        ]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=True
    )

    mock_requests.get.assert_called_with(
        "http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name/file"
    )

    assert len(fake_process.calls) == 6
    assert Exact("Installing snap 'test-name' from host (classic=True)") in logs.debug
    assert "Revisions found: host='2', target='1'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "2",
        "source": snap_installer.SNAP_SRC_HOST,
    }


def test_inject_from_host_strict(
    config_fixture,
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    logs,
    tmp_path,
):
    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "ack",
            "/tmp/test-name.assert",
        ]
    )
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
        ]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=False
    )

    mock_requests.get.assert_called_with(
        "http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name/file"
    )

    assert len(fake_process.calls) == 6
    assert Exact("Installing snap 'test-name' from host (classic=False)") in logs.debug
    assert "Revisions found: host='2', target='1'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "2",
        "source": snap_installer.SNAP_SRC_HOST,
    }


def test_inject_from_host_snap_name(
    config_fixture,
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    logs,
    tmp_path,
):
    """Inject a snap installed locally with the `--name` parameter."""
    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])
    fake_process.register_subprocess(
        ["fake-executor", "snap", "ack", "/tmp/test-name.assert"]
    )
    fake_process.register_subprocess(
        ["fake-executor", "snap", "install", "/tmp/test-name.snap"]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name_suffix", classic=False
    )

    mock_requests.get.assert_called_with(
        "http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name_suffix/file"
    )
    assert len(fake_process.calls) == 6
    assert (
        Exact(
            "Installing snap 'test-name_suffix' from host as 'test-name' in instance "
            "(classic=False)."
        )
        in logs.debug
    )
    assert "Revisions found: host='2', target='1'" in logs.debug
    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "2",
        "source": snap_installer.SNAP_SRC_HOST,
    }


@pytest.mark.parametrize("mock_get_host_snap_info", [{"revision": "x3"}], indirect=True)
def test_inject_from_host_dangerous(
    config_fixture,
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    logs,
    tmp_path,
):
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
            "--dangerous",
        ]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=False
    )

    mock_requests.get.assert_called_with(
        "http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name/file"
    )

    assert len(fake_process.calls) == 1
    assert Exact("Installing snap 'test-name' from host (classic=False)") in logs.debug
    assert "Revisions found: host='x3', target='1'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "x3",
        "source": snap_installer.SNAP_SRC_HOST,
    }


def test_inject_from_host_not_dangerous(
    config_fixture,
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    logs,
    tmp_path,
):
    fake_process.register_subprocess(
        [
            "snap",
            "known",
            "account-key",
            "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v"
            "7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
        ]
    )
    fake_process.register_subprocess(
        [
            "snap",
            "known",
            "snap-declaration",
            "snap-name=test-name",
        ]
    )
    fake_process.register_subprocess(
        [
            "snap",
            "known",
            "snap-revision",
            "snap-revision=2",
            "snap-id=3",
        ]
    )
    fake_process.register_subprocess(
        [
            "snap",
            "known",
            "account",
            "account-id=4",
        ]
    )
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "ack",
            "/tmp/test-name.assert",
        ]
    )
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
        ]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=False
    )

    mock_requests.get.assert_called_with(
        "http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name/file"
    )

    assert len(fake_process.calls) == 6
    assert Exact("Installing snap 'test-name' from host (classic=False)") in logs.debug
    assert "Revisions found: host='2', target='1'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "2",
        "source": snap_installer.SNAP_SRC_HOST,
    }


@pytest.mark.parametrize("config_fixture", ["10"], indirect=True)
@pytest.mark.parametrize("mock_get_host_snap_info", [{"revision": "10"}], indirect=True)
def test_inject_from_host_matching_revision_no_op(
    config_fixture,
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    logs,
):
    """Injection shouldn't occur if target revision equals the host revision"""
    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=True
    )

    assert mock_requests.mock_calls == []
    assert len(fake_process.calls) == 0
    assert "Revisions found: host='10', target='10'" in logs.debug
    assert (
        "Skipping snap injection: target is already up-to-date with revision on host"
    ) in logs.debug


@pytest.mark.parametrize(
    "mock_get_host_snap_info",
    [snap_installer.SnapInstallationError("msg")],
    indirect=True,
)
def test_inject_from_host_no_snapd(mock_get_host_snap_info, fake_executor):
    """The host does not have snapd at all."""
    with pytest.raises(snap_installer.SnapInstallationError):
        snap_installer.inject_from_host(
            executor=fake_executor, snap_name="test-name", classic=False
        )


def test_inject_from_host_push_error(mock_requests, fake_executor):
    mock_executor = mock.Mock(spec=fake_executor, wraps=fake_executor)
    mock_executor.push_file.side_effect = ProviderError(brief="foo")

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer.inject_from_host(
            executor=mock_executor, snap_name="test-name", classic=False
        )

    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="failed to copy snap file for snap 'test-name'",
        details="error copying snap file into target environment",
    )
    assert exc_info.value.__cause__ is not None


def test_inject_from_host_snapd_connection_error_using_pack_fallback(
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    tmpdir,
):
    mock_requests.get.side_effect = requests.exceptions.ConnectionError()

    fake_process.register_subprocess(
        [
            "snap",
            "pack",
            "/snap/test-name/current/",
            f'--filename={pathlib.PurePosixPath(tmpdir / "test-name.snap")}',
        ]
    )
    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "ack",
            "/tmp/test-name.assert",
        ]
    )
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
        ]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=False
    )

    assert mock_requests.mock_calls == [
        mock.call.get("http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name/file"),
    ]
    assert len(fake_process.calls) == 7


def test_inject_from_host_snapd_http_error_using_pack_fallback(
    mock_get_host_snap_info,
    mock_requests,
    fake_executor,
    fake_process,
    tmpdir,
):
    mock_requests.get.return_value.raise_for_status.side_effect = (
        requests.exceptions.HTTPError()
    )
    fake_process.register_subprocess(
        [
            "snap",
            "pack",
            "/snap/test-name/current/",
            f'--filename={pathlib.PurePosixPath(tmpdir / "test-name.snap")}',
        ]
    )
    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "ack",
            "/tmp/test-name.assert",
        ]
    )
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
        ]
    )

    snap_installer.inject_from_host(
        executor=fake_executor, snap_name="test-name", classic=False
    )

    assert mock_requests.mock_calls == [
        mock.call.get("http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-name/file"),
        mock.call.get().raise_for_status(),
    ]

    assert len(fake_process.calls) == 7


def test_inject_from_host_install_failure(mock_requests, fake_executor, fake_process):
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "/tmp/test-name.snap",
            "--dangerous",
        ],
        stdout="snap error",
        stderr="snap error details",
        returncode=1,
    )

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer.inject_from_host(
            executor=fake_executor, snap_name="test-name", classic=False
        )

    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="failed to install snap 'test-name'",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )

    assert len(fake_process.calls) == 1


@pytest.mark.parametrize(
    "mock_get_snap_revision_ensuring_source", [None], indirect=True
)
def test_install_from_store_strict(
    config_fixture,
    mock_get_snap_revision_ensuring_source,
    fake_executor,
    fake_process,
    logs,
    mock_get_target_snap_revision_from_snapd,
):
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "test-name",
            "--channel",
            "test-chan",
        ],
    )

    snap_installer.install_from_store(
        executor=fake_executor,
        snap_name="test-name_suffix",
        classic=False,
        channel="test-chan",
    )

    assert len(fake_process.calls) == 1
    assert (
        Exact(
            "Installing snap 'test-name_suffix' as 'test-name' from store "
            "(channel='test-chan', classic=False).",
        )
        in logs.debug
    )
    assert "Revision found in target: None" in logs.debug
    assert "Revision after install/refresh: '4'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "4",
        "source": snap_installer.SNAP_SRC_STORE,
    }


@pytest.mark.parametrize(
    "mock_get_snap_revision_ensuring_source", [None], indirect=True
)
def test_install_from_store_classic(
    config_fixture,
    mock_get_snap_revision_ensuring_source,
    fake_executor,
    fake_process,
    logs,
    mock_get_target_snap_revision_from_snapd,
):
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "test-name",
            "--channel",
            "test-chan",
            "--classic",
        ],
    )

    snap_installer.install_from_store(
        executor=fake_executor, snap_name="test-name", classic=True, channel="test-chan"
    )

    assert len(fake_process.calls) == 1
    assert (
        Exact(
            "Installing snap 'test-name' from store "
            "(channel='test-chan', classic=True)."
        )
        in logs.debug
    )
    assert "Revision found in target: None" in logs.debug
    assert "Revision after install/refresh: '4'" in logs.debug


def test_refresh_from_store(
    config_fixture,
    mock_get_snap_revision_ensuring_source,
    fake_executor,
    fake_process,
    logs,
    mock_get_target_snap_revision_from_snapd,
):
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "refresh",
            "test-name",
            "--channel",
            "test-chan",
        ],
    )

    snap_installer.install_from_store(
        executor=fake_executor,
        snap_name="test-name",
        classic=False,
        channel="test-chan",
    )

    assert len(fake_process.calls) == 1
    assert (
        Exact(
            "Installing snap 'test-name' from store"
            " (channel='test-chan', classic=False)."
        )
        in logs.debug
    )
    assert "Revision found in target: '3'" in logs.debug
    assert "Revision after install/refresh: '4'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "4",
        "source": snap_installer.SNAP_SRC_STORE,
    }


def test_install_from_store_failure(
    config_fixture, mock_get_snap_revision_ensuring_source, fake_executor, fake_process
):
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "refresh",
            "test-name",
            "--channel",
            "test-chan",
        ],
        stdout="snap error",
        stderr="snap error details",
        returncode=1,
    )

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer.install_from_store(
            executor=fake_executor,
            snap_name="test-name",
            classic=True,
            channel="test-chan",
        )

    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="Failed to install/refresh snap 'test-name'.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )


@pytest.mark.parametrize(
    "mock_get_snap_revision_ensuring_source", [None], indirect=True
)
def test_install_from_store_trim_suffix(
    config_fixture,
    mock_get_snap_revision_ensuring_source,
    fake_executor,
    fake_process,
    logs,
    mock_get_target_snap_revision_from_snapd,
):
    """Trim the `_name` suffix from the snap name, if present."""
    fake_process.register_subprocess(
        [
            "fake-executor",
            "snap",
            "install",
            "test-name",
            "--channel",
            "test-chan",
        ],
    )

    snap_installer.install_from_store(
        executor=fake_executor,
        snap_name="test-name",
        classic=False,
        channel="test-chan",
    )

    assert len(fake_process.calls) == 1
    assert (
        Exact(
            "Installing snap 'test-name' from store"
            " (channel='test-chan', classic=False)."
        )
        in logs.debug
    )
    assert "Revision found in target: None" in logs.debug
    assert "Revision after install/refresh: '4'" in logs.debug

    # check saved config
    (saved_config_record,) = (
        x
        for x in fake_executor.records_of_push_file_io
        if "craft-instance.conf" in x["destination"]
    )
    config = InstanceConfiguration(**yaml.safe_load(saved_config_record["content"]))
    assert config.snaps is not None
    assert config.snaps["test-name"] == {
        "revision": "4",
        "source": snap_installer.SNAP_SRC_STORE,
    }


# -- tests for the helping functions


def test_get_host_snap_info_ok(responses):
    """Revision retrieved ok."""
    snap_info = {"result": {"revision": "15"}}
    responses.add(
        responses.GET,
        "http+unix://%2Frun%2Fsnapd.socket/v2/snaps/test-snap",
        json=snap_info,
    )
    result = snap_installer.get_host_snap_info("test-snap")["revision"]
    assert result == "15"


def test_get_host_snap_info_connection_error(responses):
    """Error when connecting to snapd.

    Note that nothing is added to responses, so ConnectionError will be raised when
    trying to connect, which is the effect we want.
    """
    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer.get_host_snap_info("test-snap")
    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="Unable to connect to snapd service."
    )
    assert exc_info.value.__cause__ is not None


def test_get_assertion_connection_error(mocker):
    """Raise SnapInstallation when 'snap known' call fails."""
    mocker.patch(
        "subprocess.run",
        side_effect=subprocess.CalledProcessError(
            100, ["error"], output="snap error", stderr="snap error details"
        ),
    )

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer._get_assertion(["test1", "test2"])
    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="failed to get assertions for snap",
        details="* Command that failed: 'error'\n* Command exit code: 100\n* Command output: 'snap error'\n* Command standard error output: 'snap error details'",
    )
    assert exc_info.value.__cause__ is not None


def test_add_assertions_from_host_error_on_push(
    fake_executor, fake_process, mock_requests, mocker, tmpdir
):
    """Raise SnapInstallationError when assert file cannot be pushed."""
    mock_executor = mock.Mock(spec=fake_executor, wraps=fake_executor)
    mock_executor.push_file.side_effect = ProviderError(brief="foo")

    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer._add_assertions_from_host(
            executor=mock_executor,
            snap_name="test-name",
        )

    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="failed to copy assert file for snap 'test-name'",
        details="error copying snap assert file into target environment",
    )
    assert exc_info.value.__cause__ is not None
    assert len(fake_process.calls) == 4


def test_add_assertions_from_host_error_on_ack(
    fake_executor, fake_process, mock_requests
):
    """Raise SnapInstallationError when 'snap ack' fails."""
    fake_process.register_subprocess(
        ["fake-executor", "snap", "ack", "/tmp/test-name.assert"],
        stdout="snap error",
        stderr="snap error details",
        returncode=1,
    )

    # register 'snap known' calls
    for _ in range(4):
        fake_process.register_subprocess(["snap", "known", fake_process.any()])

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer._add_assertions_from_host(
            executor=fake_executor, snap_name="test-name"
        )

    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="failed to add assertions for snap 'test-name'",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )

    assert len(fake_process.calls) == 5


def test_get_target_snap_revision_from_snapd_process_error(fake_process, fake_executor):
    """Error when running curl to get info from snapd in target environment."""
    expected_cmd = [
        "fake-executor",
        "curl",
        "--silent",
        "--unix-socket",
        "/run/snapd.socket",
        "http://localhost/v2/snaps/test-snap",
    ]
    fake_process.register_subprocess(
        expected_cmd, stdout="snap error", stderr="snap error details", returncode=1
    )

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer._get_target_snap_revision_from_snapd("test-snap", fake_executor)
    assert exc_info.value == snap_installer.SnapInstallationError(
        "Unable to get target snap revision.",
        details="* Command that failed: 'fake-executor curl --silent --unix-socket /run/snapd.socket http://localhost/v2/snaps/test-snap'\n* Command exit code: 1\n* Command output: b'snap error'\n* Command standard error output: b'snap error details'",
    )
    assert exc_info.value.__cause__ is not None


def test_get_target_snap_revision_from_snapd_ok(fake_process, fake_executor):
    """Proper info retrieved ok."""
    expected_cmd = [
        "fake-executor",
        "curl",
        "--silent",
        "--unix-socket",
        "/run/snapd.socket",
        "http://localhost/v2/snaps/test-snap",
    ]
    fake_snapd_response = json.dumps({"status-code": 200, "result": {"revision": "17"}})
    fake_process.register_subprocess(expected_cmd, stdout=fake_snapd_response)

    result = snap_installer._get_target_snap_revision_from_snapd(
        "test-snap", fake_executor
    )
    assert result == "17"


def test_get_target_snap_revision_from_snapd_not_found(fake_process, fake_executor):
    """The requested snap is not found in target's snapd."""
    expected_cmd = [
        "fake-executor",
        "curl",
        "--silent",
        "--unix-socket",
        "/run/snapd.socket",
        "http://localhost/v2/snaps/test-snap",
    ]
    fake_snapd_response = {"status-code": 404}
    fake_process.register_subprocess(
        expected_cmd, stdout=json.dumps(fake_snapd_response)
    )

    result = snap_installer._get_target_snap_revision_from_snapd(
        "test-snap", fake_executor
    )
    assert result is None


def test_get_target_snap_revision_from_snapd_unknown(fake_process, fake_executor):
    """The target's snapd returns an unknown response."""
    expected_cmd = [
        "fake-executor",
        "curl",
        "--silent",
        "--unix-socket",
        "/run/snapd.socket",
        "http://localhost/v2/snaps/test-snap",
    ]
    fake_snapd_response = {"status-code": 666}
    fake_process.register_subprocess(
        expected_cmd, stdout=json.dumps(fake_snapd_response)
    )

    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer._get_target_snap_revision_from_snapd("test-snap", fake_executor)
    assert exc_info.value == snap_installer.SnapInstallationError(
        f"Unknown response from snapd: {fake_snapd_response!r}"
    )


def test_get_snap_revision_ensuring_source_ok(config_fixture, fake_executor):
    """Snap is available being installed by specified source."""
    result = snap_installer._get_snap_revision_ensuring_source(
        "test-name", snap_installer.SNAP_SRC_HOST, fake_executor
    )
    assert result == "1"


@pytest.mark.parametrize(
    "fake_config",
    [
        None,
        InstanceConfiguration(snaps=None),
        InstanceConfiguration(snaps={"othersnap": {"revision": "1000"}}),
    ],
)
def test_get_snap_revision_ensuring_source_no_config(fake_config, fake_executor):
    """No config available for the indicated snap."""
    with mock.patch.object(InstanceConfiguration, "load", return_value=fake_config):
        result = snap_installer._get_snap_revision_ensuring_source(
            "test-name", snap_installer.SNAP_SRC_HOST, fake_executor
        )
    assert result is None


def test_get_snap_revision_ensuring_source_missing_source(
    fake_process, fake_executor, config_fixture
):
    """Support for instances that have old configuration and then the lib is updated."""
    fake_config = InstanceConfiguration(snaps={"test-name": {"revision": "1000"}})
    fake_process.register_subprocess(["fake-executor", "snap", "remove", "test-name"])
    with mock.patch.object(InstanceConfiguration, "load", return_value=fake_config):
        result = snap_installer._get_snap_revision_ensuring_source(
            "test-name", snap_installer.SNAP_SRC_STORE, fake_executor
        )
    assert result is None
    assert len(fake_process.calls) == 1


def test_get_snap_revision_ensuring_source_different_source_ok(
    fake_process, fake_executor, config_fixture
):
    """Snap is available but from other source; snap removal ended ok."""
    fake_process.register_subprocess(["fake-executor", "snap", "remove", "test-name"])
    result = snap_installer._get_snap_revision_ensuring_source(
        "test-name", snap_installer.SNAP_SRC_STORE, fake_executor
    )
    assert result is None
    assert len(fake_process.calls) == 1


def test_get_snap_revision_ensuring_source_different_source_error(
    fake_process, fake_executor, config_fixture
):
    """Snap is available but from other source; snap removal failed."""
    fake_process.register_subprocess(
        ["fake-executor", "snap", "remove", "test-name"],
        stdout="snap error",
        stderr="snap error details",
        returncode=1,
    )
    with pytest.raises(snap_installer.SnapInstallationError) as exc_info:
        snap_installer._get_snap_revision_ensuring_source(
            "test-name", snap_installer.SNAP_SRC_STORE, fake_executor
        )
    assert exc_info.value == snap_installer.SnapInstallationError(
        brief="Failed to remove snap 'test-name'.",
        details=details_from_called_process_error(
            exc_info.value.__cause__  # type: ignore
        ),
    )
    assert exc_info.value.__cause__ is not None


# -- tests for the Snap pydantic model


def test_snaps_no_channel_raises_errors(fake_executor):
    """Verify the Snap model raises an error when the channel is an empty string."""
    with pytest.raises(BaseConfigurationError) as exc_info:
        Snap(name="snap1", channel="")

    assert exc_info.value == BaseConfigurationError(
        brief="channel cannot be empty",
        resolution="set channel to a non-empty string or `None`",
    )



#
# Copyright 2021 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


from craft_providers.multipass import errors


def test_multipass_error():
    error = errors.MultipassError(brief="foo")

    assert str(error) == "foo"


def test_multipass_error_with_details():
    error = errors.MultipassError(brief="foo", details="bar")

    assert str(error) == "foo\nbar"


def test_multipass_error_with_details_resolution():
    error = errors.MultipassError(brief="foo", details="bar", resolution="do this")

    assert str(error) == "foo\nbar\ndo this"


def test_multipass_installation_error():
    error = errors.MultipassInstallationError(reason="error during foo")

    assert str(error) == (
        "Failed to install Multipass: error during foo.\n"
        "Please visit https://multipass.run/ for instructions"
        " on installing Multipass for your operating system."
    )


def test_multipass_installation_error_with_details():
    error = errors.MultipassInstallationError(
        reason="error during foo", details="Some details..."
    )

    assert str(error) == (
        "Failed to install Multipass: error during foo.\n"
        "Some details...\n"
        "Please visit https://multipass.run/ for instructions"
        " on installing Multipass for your operating system."
    )



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from unittest import mock

import pytest
from craft_providers import Base, bases, multipass


@pytest.fixture()
def mock_base_configuration():
    return mock.Mock(spec=Base)


@pytest.fixture()
def mock_multipass_instance():
    with mock.patch(
        "craft_providers.multipass._launch.MultipassInstance",
        spec=multipass.MultipassInstance,
    ) as mock_instance:
        mock_instance.return_value.name = "test-instance"
        yield mock_instance.return_value


def test_launch_fresh(mock_base_configuration, mock_multipass_instance):
    mock_multipass_instance.exists.return_value = False

    multipass.launch(
        "test-instance", base_configuration=mock_base_configuration, image_name="30.04"
    )

    assert mock_multipass_instance.mock_calls == [
        mock.call.exists(),
        mock.call.launch(cpus=2, disk_gb=64, mem_gb=2, image="30.04"),
    ]
    assert mock_base_configuration.mock_calls == [
        mock.call.setup(executor=mock_multipass_instance)
    ]


def test_launch_with_existing_instance(
    mock_base_configuration, mock_multipass_instance
):
    mock_multipass_instance.exists.return_value = True

    multipass.launch(
        "test-instance", base_configuration=mock_base_configuration, image_name="30.04"
    )

    assert mock_multipass_instance.mock_calls == [
        mock.call.exists(),
        mock.call.start(),
    ]
    assert mock_base_configuration.mock_calls == [
        mock.call.warmup(executor=mock_multipass_instance)
    ]


def test_launch_with_existing_instance_incompatible_with_auto_clean(
    mock_base_configuration, mock_multipass_instance
):
    mock_multipass_instance.exists.return_value = True
    mock_base_configuration.warmup.side_effect = [
        bases.BaseCompatibilityError(reason="foo"),
        None,
    ]

    multipass.launch(
        "test-instance",
        base_configuration=mock_base_configuration,
        cpus=2,
        disk_gb=64,
        mem_gb=2,
        image_name="30.04",
        auto_clean=True,
    )

    assert mock_multipass_instance.mock_calls == [
        mock.call.exists(),
        mock.call.start(),
        mock.call.delete(),
        mock.call.launch(cpus=2, disk_gb=64, mem_gb=2, image="30.04"),
    ]
    assert mock_base_configuration.mock_calls == [
        mock.call.warmup(executor=mock_multipass_instance),
        mock.call.setup(executor=mock_multipass_instance),
    ]


def test_launch_with_existing_instance_incompatible_without_auto_clean(
    mock_base_configuration, mock_multipass_instance
):
    mock_multipass_instance.exists.return_value = True
    mock_base_configuration.warmup.side_effect = [
        bases.BaseCompatibilityError(reason="foo")
    ]

    with pytest.raises(bases.BaseCompatibilityError):
        multipass.launch(
            "test-instance",
            base_configuration=mock_base_configuration,
            image_name="30.04",
            auto_clean=False,
        )

    assert mock_multipass_instance.mock_calls == [
        mock.call.exists(),
        mock.call.start(),
    ]
    assert mock_base_configuration.mock_calls == [
        mock.call.warmup(executor=mock_multipass_instance)
    ]



#
# Copyright 2021-2022 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import contextlib
import io
import pathlib
import subprocess
from typing import Any, Dict, List, Optional

import pytest
import responses as responses_module
from craft_providers.executor import Executor
from craft_providers.util import env_cmd

# command used by the FakeExecutor
DEFAULT_FAKE_CMD = ["fake-executor"]


class FakeExecutor(Executor):
    """Fake Executor.

    Provides a fake execution environment meant to be paired with the
    fake_subprocess fixture for complete control over execution behaviors.

    Calls to each method are recorded in `records_of_<method_name>` for introspection,
    similar to mock_calls.
    """

    def __init__(self) -> None:
        self.records_of_push_file_io: List[Dict[str, Any]] = []
        self.records_of_pull_file: List[Dict[str, Any]] = []
        self.records_of_push_file: List[Dict[str, Any]] = []
        self.records_of_delete: List[Dict[str, Any]] = []
        self.records_of_exists: List[Dict[str, Any]] = []
        self.records_of_mount: List[Dict[str, Any]] = []
        self.records_of_is_running: List[Dict[str, Any]] = []

    def push_file_io(
        self,
        *,
        destination: pathlib.PurePath,
        content: io.BytesIO,
        file_mode: str,
        group: str = "root",
        user: str = "root",
    ) -> None:
        self.records_of_push_file_io.append(
            {
                "destination": destination.as_posix(),
                "content": content.read(),
                "file_mode": file_mode,
                "group": group,
                "user": user,
            }
        )

    def execute_popen(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.Path] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        **kwargs,
    ) -> subprocess.Popen:
        env_args = [] if env is None else env_cmd.formulate_command(env, chdir=cwd)

        final_cmd = [*DEFAULT_FAKE_CMD, *env_args, *command]
        return subprocess.Popen(final_cmd, **kwargs)

    def execute_run(
        self,
        command: List[str],
        *,
        cwd: Optional[pathlib.Path] = None,
        env: Optional[Dict[str, Optional[str]]] = None,
        timeout: Optional[float] = None,
        check: bool = False,
        **kwargs,
    ) -> subprocess.CompletedProcess:
        env_args = [] if env is None else env_cmd.formulate_command(env, chdir=cwd)

        final_cmd = [*DEFAULT_FAKE_CMD, *env_args, *command]

        return subprocess.run(final_cmd, timeout=timeout, check=check, **kwargs)

    def pull_file(self, *, source: pathlib.PurePath, destination: pathlib.Path) -> None:
        self.records_of_pull_file.append(
            {
                "source": source,
                "destination": destination,
            }
        )

    def push_file(self, *, source: pathlib.Path, destination: pathlib.PurePath) -> None:
        self.records_of_push_file.append(
            {
                "source": source,
                "destination": destination,
            }
        )

    def delete(self) -> None:
        self.records_of_delete.append({})

    def exists(self) -> bool:
        self.records_of_exists.append({})
        return True

    def mount(self, *, host_source: pathlib.Path, target: pathlib.Path) -> None:
        self.records_of_mount.append({"host_source": host_source, "target": target})

    def is_running(self) -> bool:
        self.records_of_is_running.append({})
        return True


@pytest.fixture()
def fake_executor():
    return FakeExecutor()


@pytest.fixture()
def responses():
    """Simple helper to use responses module as a fixture.

    Used for easier integration in tests.
    """
    with responses_module.RequestsMock() as rsps:
        yield rsps


@pytest.fixture(
    params=range(4), ids=("succeed", "fail_once", "fail_twice", "fail_thrice")
)
def failure_count(request):
    return request.param


@pytest.fixture(autouse=True)
def fake_home_temporary_directory(monkeypatch, tmp_path):
    @contextlib.contextmanager
    def home_temporary_directory():
        yield tmp_path

    monkeypatch.setattr(
        "craft_providers.instance_config.temp_paths.home_temporary_directory",
        home_temporary_directory,
    )
    return tmp_path


@pytest.fixture(autouse=True)
def fake_home_temporary_file(monkeypatch, tmp_path):
    temp_file = tmp_path / "temp-file"

    @contextlib.contextmanager
    def home_temporary_file():
        temp_file.touch()
        yield temp_file
        temp_file.unlink(missing_ok=True)

    monkeypatch.setattr(
        "craft_providers.instance_config.temp_paths.home_temporary_file",
        home_temporary_file,
    )
    return temp_file


@pytest.fixture()
def stub_verify_network(fake_process):
    """Ensures network check for Executor.execute_run(verify_network=True) succeeds."""
    fake_process.register_subprocess(
        [*DEFAULT_FAKE_CMD, "bash", "-c", "exec 3<> /dev/tcp/snapcraft.io/443"],
        returncode=0,
    )
